# Copyright 2023 Agnostiq Inc.

"""Functions to convert tg -> TransportGraphSchema"""

import os
from pathlib import Path
from typing import List

import networkx as nx
from covalent._workflow.transport import _TransportGraph

from ..shared.schemas.edge import EdgeMetadata, EdgeSchema
from ..shared.schemas.electron import ElectronSchema
from ..shared.schemas.transport_graph import TransportGraphSchema
from .electron import deserialize_node, serialize_node


def serialize_edge(source: int, target: int, attrs: dict) -> EdgeSchema:
    meta = EdgeMetadata(
        edge_name=attrs["edge_name"],
        param_type=attrs.get("param_type"),
        arg_index=attrs.get("arg_index"),
    )
    return EdgeSchema(source=source, target=target, metadata=meta)


def deserialize_edge(e: EdgeSchema) -> dict:
    return {
        "source": e.source,
        "target": e.target,
        "attrs": e.metadata.dict(),
    }


def _serialize_nodes(g: nx.MultiDiGraph, storage_path: str) -> List[ElectronSchema]:
    results = []
    base_path = Path(storage_path)
    for i in g.nodes:
        node_storage_path = base_path / f"node_{i}"
        os.mkdir(node_storage_path)
        results.append(serialize_node(i, g.nodes[i], node_storage_path))
    return results


def _serialize_edges(g: nx.MultiDiGraph) -> List[EdgeSchema]:
    results = []
    for edge in g.edges:
        source, target, key = edge
        results.append(serialize_edge(source, target, g.edges[edge]))
    return results


def serialize_transport_graph(tg, storage_path: str) -> TransportGraphSchema:
    g = tg.get_internal_graph_copy()
    return TransportGraphSchema(
        nodes=_serialize_nodes(g, storage_path),
        links=_serialize_edges(g),
    )


def deserialize_transport_graph(t: TransportGraphSchema) -> _TransportGraph:
    tg = _TransportGraph()
    g = tg._graph
    nodes = [deserialize_node(n) for n in t.nodes]
    edges = [deserialize_edge(e) for e in t.links]
    for node in nodes:
        node_id = node["id"]
        attrs = node["attrs"]
        g.add_node(node_id, **attrs)
    for edge in edges:
        x = edge["source"]
        y = edge["target"]
        g.add_edge(x, y, **edge["attrs"])

    return tg



# Copyright 2023 Agnostiq Inc.


"""Models for /api/v1/resultv2 endpoints"""

from typing import Optional

from pydantic import BaseModel

EDGE_METADATA_KEYS = {
    "edge_name",
    "param_type",
    "arg_index",
}


class EdgeMetadata(BaseModel):
    edge_name: str
    param_type: Optional[str]
    arg_index: Optional[int]


class EdgeSchema(BaseModel):
    source: int
    target: int
    metadata: EdgeMetadata



# Copyright 2023 Agnostiq Inc.

"""Models for /api/v1/resultv2 endpoints"""

from typing import Optional

from pydantic import BaseModel


class AssetSchema(BaseModel):
    digest_alg: Optional[str]
    digest: Optional[str]
    uri: Optional[str]
    remote_uri: Optional[str]

    # Size of the asset in bytes
    size: Optional[int]



# Copyright 2023 Agnostiq Inc.


import functools

from typing import Callable, List, Optional, Union

import pennylane as qml

from ..core.qnode_qe import QElectronInfo, QNodeQE
from ..executors.base import CloudQCluster, CloudQExecutor

Selector = Union[str, Callable[[qml.tape.QuantumScript, List[CloudQExecutor]], CloudQExecutor]]


def qelectron(
    qnode: Optional[qml.QNode] = None,
    *,
    executors: Union[CloudQExecutor, CloudQCluster, List[CloudQExecutor]] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    selector: str = "cyclic",
) -> QNodeQE:
    """
    QElectron decorator to be called upon a Pennylane QNode. Adds multi-backend
    execution functionality to the original QNode.

    Args:
        qnode: The Pennylane :code:`QNode` to wrap.

    Keyword Args:
        executors: The quantum executor(s) to use for running the QNode. A single
            executor, list of executors, or a :code:`QCluster` instance are accepted.
            If a list of multiple executors is passed, a quantum cluster is
            initialized from this list automatically and :code:`selector` is used as the
            cluster's selector. Defaults to a :code:`CloudQExecutor`.
        name: An optional name for the QElectron. Defaults to the circuit function's
            name.
        description: An optional description for the QElectron. Defaults to the
            circuit function's docstring.
        selector: One of the strings :code:`"cyclic"` or :code:`"random"`. Used to
            select the executor from the cluster when multiple executors are passed.

    Raises:
        ValueError: If any invalid executors are passed.

    Returns:
        :code:`QNodeQE`: A sub-type of :code:`QNode` that integrates QElectrons.
    """

    if executors is None:
        executors = CloudQExecutor()

    # Check if executor is a list of executors.
    if isinstance(executors, list):
        if not all(isinstance(ex, CloudQExecutor) for ex in executors):
            raise ValueError("Invalid executor in executors list.")
        if len(executors) > 1:
            # Convert to cluster if more than one executor in list.
            executors = CloudQCluster(executors=executors, selector=selector)

    # Set default name and description.
    if name is None:
        name = qnode.func.__name__

    if description is None:
        description = qnode.func.__doc__

    qelectron_info = QElectronInfo(
        name=name,
        description=description,
        qnode_device_shots=qnode.device.shots,
        num_device_wires=qnode.device.num_wires,
        pennylane_active_return=qml.active_return(),
    )

    # Validating shots and wires for the batch of circuits
    # Here the `executors` will either be a CloudQExecutor or a CloudQCluster
    executors.set_shots(qelectron_info.qnode_device_shots)
    executors.validate_attrs()

    # Check if a single executor instance was passed.
    if not isinstance(executors, list):
        executors = [executors]

    if qnode is None:
        # This only happens when `qelectron()` is not used as a decorator.
        return functools.partial(
            qelectron,
            executors=executors,
            name=name,
            description=description
        )

    # Create and return a new `QNodeQE` instance.
    return QNodeQE(qnode, executors, qelectron_info)



# Copyright 2023 Agnostiq Inc.

"""FastAPI models for /api/v1/resultv2 endpoints"""

import platform

from pydantic import BaseModel

from .asset import AssetSchema
from .transport_graph import TransportGraphSchema

LATTICE_METADATA_KEYS = {
    "__name__",
    # metadata
    "executor",
    "workflow_executor",
    "executor_data",
    "workflow_executor_data",
}

LATTICE_ASSET_KEYS = {
    "workflow_function",
    "workflow_function_string",
    "__doc__",
    "named_args",
    "named_kwargs",
    "cova_imports",
    "lattice_imports",
    # metadata
    "deps",
    "call_before",
    "call_after",
}

LATTICE_FUNCTION_FILENAME = "function.pkl"
LATTICE_FUNCTION_STRING_FILENAME = "function_string.txt"
LATTICE_DOCSTRING_FILENAME = "function_docstring.txt"
LATTICE_EXECUTOR_DATA_FILENAME = "executor_data.pkl"
LATTICE_WORKFLOW_EXECUTOR_DATA_FILENAME = "workflow_executor_data.pkl"
LATTICE_ERROR_FILENAME = "error.log"
LATTICE_INPUTS_FILENAME = "inputs.pkl"
LATTICE_NAMED_ARGS_FILENAME = "named_args.pkl"
LATTICE_NAMED_KWARGS_FILENAME = "named_kwargs.pkl"
LATTICE_RESULTS_FILENAME = "results.pkl"
LATTICE_DEPS_FILENAME = "deps.pkl"
LATTICE_CALL_BEFORE_FILENAME = "call_before.pkl"
LATTICE_CALL_AFTER_FILENAME = "call_after.pkl"
LATTICE_COVA_IMPORTS_FILENAME = "cova_imports.pkl"
LATTICE_LATTICE_IMPORTS_FILENAME = "lattice_imports.pkl"
LATTICE_STORAGE_TYPE = "file"


ASSET_FILENAME_MAP = {
    "workflow_function": LATTICE_FUNCTION_FILENAME,
    "workflow_function_string": LATTICE_FUNCTION_STRING_FILENAME,
    "doc": LATTICE_DOCSTRING_FILENAME,
    "named_args": LATTICE_NAMED_ARGS_FILENAME,
    "named_kwargs": LATTICE_NAMED_KWARGS_FILENAME,
    "cova_imports": LATTICE_COVA_IMPORTS_FILENAME,
    "lattice_imports": LATTICE_LATTICE_IMPORTS_FILENAME,
    "deps": LATTICE_DEPS_FILENAME,
    "call_before": LATTICE_CALL_BEFORE_FILENAME,
    "call_after": LATTICE_CALL_AFTER_FILENAME,
}


class LatticeAssets(BaseModel):
    workflow_function: AssetSchema
    workflow_function_string: AssetSchema
    doc: AssetSchema  # __doc__
    named_args: AssetSchema
    named_kwargs: AssetSchema
    cova_imports: AssetSchema
    lattice_imports: AssetSchema

    # lattice.metadata
    deps: AssetSchema
    call_before: AssetSchema
    call_after: AssetSchema


class LatticeMetadata(BaseModel):
    name: str  # __name__
    executor: str
    executor_data: dict
    workflow_executor: str
    workflow_executor_data: dict
    python_version: str = platform.python_version()


class LatticeSchema(BaseModel):
    metadata: LatticeMetadata
    assets: LatticeAssets
    transport_graph: TransportGraphSchema



# Copyright 2023 Agnostiq Inc.


from .base_client import BaseQClient


class CloudQClient(BaseQClient):

    def submit(self, qscripts, executors, qelectron_info, qnode_specs):
        raise NotImplementedError

    def get_results(self, batch_id):
        raise NotImplementedError

    @property
    def selector(self):
        raise NotImplementedError

    @property
    def database(self):
        raise NotImplementedError

    def serialize(self, obj):
        raise NotImplementedError

    def deserialize(self, ser_obj):
        raise NotImplementedError

